/*****************************************************************************
******************************************************************************
This program is part of the RC-Crypt suite
that implements rc5 128 bit block cipher.

Copyright (C) 2001-2 R.G.Jones

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Email:rich@ricksoft.co.uk
********************************************************************************
File format:
The encoded output is an integer number of 16 byte blocks.
The first block is pseudo random data.
The second block to the last but one block are encoded data.
If the data does not fill the last but one block, it is padded 
with pseudo random data to make it sixteen bytes.
The last block is pseudo random data, apart from the lowest 
four bits of the first byte. These are the number of bytes
used to pad the last but one byte (ie 0-15).
********************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <pwd.h>
#include <time.h>
#include <arpa/inet.h>

#include "rccrypt.h"

/************************************************************/
/*Global variables used by different subroutines*/
static long rounds;	/*number of rounds required (0..255)*/
static ULLONG *S;	/*Pointer to subkey array*/

/********************************************************************************/
/* The following are definitions of subroutines defined after main*/
void crypt_it(char *buffer, BOOLEAN decode);

/********************************************************************************/
/* Main program begins here*/
/********************************************************************************/
int main(int argc, char **argv)
{
/********************************************************************************/
/*Declare all variables here*/
struct passwd *user_data;	/*will contain details of user data*/
int d;	/*used for determining command line options*/
long i;	/*used for loop control*/
int one_char;	/*used for getting chars from the input stream*/
BOOLEAN decode_option;
FILE *input;	/*either from a file or the standard input*/
FILE *output;	/*either to a file or the standard ouput*/
FILE *keyfile;	/*used to input the key if it is in a file*/
char key_buffer[MAX_BYTES_IN_KEY + 1];	/*to hold key if input from a file*/
char my_cwd[500];	/*to hold the directory the program is started in*/
UBYTE *padded_keystring;	/*Pointer to padded array for the key to fit into*/
UBYTE sub_array[9];		/*array to get chunks out of the padded array*/
ULLONG *K;	/*Pointer for integer access to key*/
long c;		/*number of 64 bit integers in the key*/
long t;		/*number of subkeys:2*(rounds+1)*/
long j,k;	/*used to access arrays*/
long num_chars_in_keystring;
long num_chars_in_padded_keystring;
long num_lls_in_key;
char *k_chars;	/*points to key string given in command line*/
char *f_chars;  /*Points to file containing key*/
char *r_chars;	/*points to rounds string given in command line*/
char *infile;	/*points to filename if given on command line*/
char *outfile;	/*points to filename if given on command line*/
char *tail;	/*for char to number conversion checking*/
ULONG temp;	/*used for VERY temporary storage of a number...*/
ULLONG A,B;	/*64-bit integers used to mangle sub keys*/
ULONG char_array[5];	/*Used to create aligned character buffer that can be printed*/
char *buffer = (char *)char_array;
					/*buffer must be aligned to a long as a type conversion is applied*/
ULONG previous_array[5];	/*Used to create aligned character buffer that can be printed*/
char *previous_block = (char *)previous_array;
					/*previous_block must be aligned to a long as a type conversion is applied*/
size_t count, rd_count;		/*how many items were reading/writing the stream*/
UBYTE pad_value;	/*Number of pad characters used when encoding*/
BOOLEAN firstblock;	/*To check if this is the first block*/
BOOLEAN secondblock;	/*To check if this is the second block*/
int to_print;		/*how much of the block to print when decoding*/

/********************************************************************************/
/*Create a random seed based on the calendar time*/
	srandom(time(NULL));

/********************************************************************************/
/*Set all variables etc to default/null values*/
	opterr = 0;
	num_chars_in_keystring=0;
	rounds=100;
	decode_option=FALSE;
	k_chars=NULL;
	f_chars=NULL;
	r_chars=NULL;
	keyfile=NULL;
	infile=NULL;
	outfile=NULL;

/********************************************************************************/
/*Check out what options have been given and set variables*/
	while ((d = getopt (argc, argv, "dk:f:r:i:o:")) != -1)
		switch (d)
		{
			case 'o':
				outfile=optarg;
				break;
			case 'i':
				infile=optarg;
				break;
			case 'd':
				decode_option=TRUE;
				break;
			case 'k':
				k_chars = optarg;
				break;
			case 'f':
				f_chars = optarg;
				break;
			case 'r':
				r_chars = optarg;
				/*set the number of rounds*/
				errno=0;
				rounds=strtol(r_chars,&tail,10);
				if ((tail==r_chars) 
					|| (errno) 
					|| (rounds>255) 
					|| (rounds<0))
					bail("ERROR: Rounds must be 0..255");
				break;
			case '?':
				/* 
				* The unknown case deliberate falls
				* through to the default case
				*/
			default:
				fprintf(stderr,"Usage:\n\t%s\t [-d] [-r rounds ] "
								"[ -k key | -f keyfile] [ -i infilename ] "
								"[ -o outfilename ]\n\n",
								argv[0]);
				fprintf(stderr,"\nSee README file for further information\n");
				return 1;
		}
		
/********************************************************************************/
/* If a key has not been given, use the file given or ~/.rccrypt_key*/
	if (NULL == k_chars)
	{
		if (NULL == f_chars)
		{
			/*remember what directory we are in*/
			if (getcwd(my_cwd,500) != NULL)
			{
				/*find the user's home directory*/
				user_data=getpwuid(getuid());
				if (user_data == NULL) 
					bail("Unable to determine user's home directory");

				if( user_data->pw_dir != NULL)
				{
					if (chdir(user_data->pw_dir)) 
						bail("Unable to change directory");
					keyfile = fopen(".rccrypt_key","r");
					/*return to the original directory*/
					if (chdir(my_cwd)) 
						bail("Unable to change directory");
				}
			}
			else
			{
				bail("Unable to determine current directory.");
			}
		}
		else
		{
			keyfile = fopen(f_chars,"r");
		}

		/*read in the key if the file exists*/
		if (keyfile != NULL)
		{
			rd_count = fread (key_buffer,1,MAX_BYTES_IN_KEY,keyfile);
			/*make the end null*/
			key_buffer[rd_count-1]='\0';
			/*set the k_chars to point to the buffer*/
			k_chars = key_buffer;
			/*close the file*/
			fclose(keyfile);
		}
		/*note - if get to here without creating key, an error will occur*/
	}
/********************************************************************************/
/*set the number of chars in the key*/
	if(k_chars != NULL)
	{
		num_chars_in_keystring=strlen(k_chars);
		if (num_chars_in_keystring > 64)
			num_chars_in_keystring = 64;
	}
	else
	{
		bail("You must supply a key");
	}

/********************************************************************************
Converting the key....
Each 2 chars in the key is one byte.
One long has 8 chars.
One Long long has 16 chars.
Create a string containing the chars in the key, padded with enough zeros
at the beginning to make it up to an integer number of long longs.
Create the final long long array the right size.
The use atoul(base 16) to convert the char array 8 chars at a time into the final array.
********************************************************************************/
	num_lls_in_key=num_chars_in_keystring/BLOCK_SIZE;

	if (remainder(num_chars_in_keystring,16)>0)
		num_lls_in_key +=1;

	c = num_lls_in_key;

	num_chars_in_padded_keystring = num_lls_in_key*BLOCK_SIZE;

/*create the padded array*/

	padded_keystring = (UBYTE *)malloc(num_chars_in_padded_keystring+1);
	if (NULL == padded_keystring) 
		bail("Malloc error:padded_keystring");

/*fill the array with zeros then copy the key to the end of it*/

	memset(padded_keystring, '0', num_chars_in_padded_keystring);
	
/*make sure it finishes with a null char so can use string fns on it*/
	padded_keystring[num_chars_in_padded_keystring] = '\0';
	strncpy(padded_keystring + 
			num_chars_in_padded_keystring -
			num_chars_in_keystring,
			k_chars,
			num_chars_in_keystring);

/*create the long long array for the key*/#

	K = (ULLONG *) malloc( num_lls_in_key * 8 );
	if ( NULL == K) 
		bail("Malloc error:K");

/*fill it with zeros*/

	memset( K, 0, num_lls_in_key * 8);

/*now convert each chunk of 8 chars to a long and copy it into the array*/

	for (i=0,j=0; i<num_chars_in_padded_keystring; i+=8,j+=4)
	{
		strncpy(sub_array,padded_keystring+i,8);
		sub_array[8]='\0';

		errno = 0;
		temp = strtoul(sub_array, &tail, BLOCK_SIZE);
		if (tail != (char *)(sub_array+8)) 
			bail("Key must be a hex number");

		/*Put key values in network order*/
		temp = htonl( temp );

		if(errno) 
			bail("Key: overflow on conversion");
			
		/*copy the ULONG into the string*/
		memcpy(((char *)K)+j, (char *)&temp, 4);
	}

	
/********************************************************************************/
/*Now fill the subkeys with pseudo random bits*/

	t = 2 * (rounds+1);
	
	S = (ULLONG *)malloc(t*8);
	if (NULL == S) 
		bail("Malloc Error:S");
		
	S[0] = P64;
	for(i=1; i<t; i++)
	{
		S[i] =S [i-1] + Q64;
	}

	/*Convert subkeys to network byte order*/
	for (i=0; i<t; i++)
	{
		S[i] = htonll(S[i]);
		uprint ("Subkey",S[i]);	
	}

	for (i=0; i<c; i++)
	{
		uprint("Host order key",ntohll(K[i]));
		uprint("Network order key",K[i]);	
	}
		
	/*Mangle the subkeys with the Key*/
	k = j = 0;
	A = B = 0;
	for(i=0; i < 3*max(t,c) ;i++)
	{
	/*At this point, both K and S are in network byte order*/
		S[j] = htonll(lrotate( ntohll(S[j]) + A + B, 3));
		K[k] = htonll(lrotate( ntohll(K[k]) + A + B, A + B));
		A = ntohll(S[j]);
		B = ntohll(K[k]);
		if (++j >= t) j=0;
		if (++k >= c) k=0;
	}
	
	/*Print out mangled keys*/
	for (i=0; i<t; i++)
	{
		uprint("Mangled Subkey=",S[i]);
	}
	
/********************************************************************************/
/*Determine where our input will be from*/
	if (infile == NULL)
	{
		input=stdin;/*get info from the standard input*/
		/*this allows a user to pipe stuff in...*/
	}
	else
	{
		input=fopen(infile,"r");
		if(input==NULL) 
			bail("Failed to open input file");
	}
	
/********************************************************************************/
/*Determine where the output goes to*/
	if (NULL == outfile)
	{
		output = stdout; /*send output to the standard output*/
		/*allows a user to pipe or redirect stuff out*/
	}
	else
	{
		output = fopen(outfile,"w");
		if (NULL == output) 
			bail("Failed to open output file");
	}
	
/********************************************************************************
Now we can do some encryption
We need to get 2 64-bit words for each block to encrypt
if there is not enough data, we need to pad the words with extra bytes
this program uses random printable chars for padding
********************************************************************************/

	pad_value = 0;
	firstblock = TRUE;
	secondblock=FALSE;
	
	one_char=getc(input);

	if ((firstblock) && !decode_option)
	{
		/*
		* When encoding;
		* Output some random data for the first block.
		* This will be used in a future release
		* so we can xor each block of data with a previous block
		*/
		/*Fill the buffer with random data*/	
		for (i=0; i<BLOCK_SIZE; i++)
			buffer[i] = get_random_char();

		/*Do the same encoding as before*/
		crypt_it(buffer, decode_option);

		/*now output the data we worked on*/
		count=fwrite(buffer, 1, BLOCK_SIZE, output);
		if (count != BLOCK_SIZE)
			bail("Error in writing to output stream-first block\n");

		firstblock=FALSE;
	}
				
	while(one_char!=EOF)
	{
		ungetc(one_char,input);/*put back the character as it isn't an eof*/
		/*
		* Read BLOCK_SIZE bytes of data into the buffer
		* If we don't get BLOCK_SIZE bytes, pad the buffer with 
		* random printable characters to overwrite previous values.
		* When decoding, we will always have BLOCK_SIZE bytes.
		*/
		rd_count = fread(buffer, 1, BLOCK_SIZE, input);
		
		if (rd_count < BLOCK_SIZE)
		{
			pad_value = BLOCK_SIZE - rd_count;
			for (i=rd_count; i<BLOCK_SIZE; i++)
				buffer[i] = get_random_char();
		}

		/*Do the encryption/decryption*/
		crypt_it (buffer, decode_option);
		
		/*Get the next char so we can test it for eof*/
		one_char=getc(input);
			
		/*Decide whether to print out all the data or not*/
		if (TRUE == decode_option)
		{
			/*
			* How many bytes should we print from
			* the previous block?
			*/
			if (EOF == one_char) /*if true, we are on the last  block*/
			{
				/*
				* Get the pad value
				* buffer should be in network byte order
				*/	
				uprint("Buffer",*buffer);			
				memcpy (&pad_value, buffer, 1);
				
				/*throw away the top 4 bits*/
				pad_value &= 0xf;
				to_print = BLOCK_SIZE - pad_value;
			}
			else
			{
				to_print = BLOCK_SIZE;
			}
			
			/*We only start printing output when we get to the third block*/
			if ((!firstblock) && (!secondblock))
			{
				count = fwrite (previous_block, 1, to_print, output);
				if ( count != to_print )
					bail("Error in writing to output stream-not firstblock");
			}
			else
			{
				firstblock = FALSE;
				if (secondblock)
					secondblock = FALSE;
				else
					secondblock = TRUE;
			}
			
		}
		else /*encoding*/
		{
			/*output the data we encoded*/
			count = fwrite (buffer, 1, BLOCK_SIZE, output);
			if (count != BLOCK_SIZE)
				bail("Error in writing to output stream-encoding");
			
			if (EOF == one_char) /*if true, we are on the last  block*/
			{
				/*Output the padding number*/
				/*Fill the buffer with random data*/	
				for (i=0; i<BLOCK_SIZE; i++)
					buffer[i] = get_random_char();
				
				/*Put the pad size into the correct bits of the first byte*/
				pad_value &= 0xf;	/*lower 4 bits*/
				*buffer &= 0xf0;	/*upper 4 bits*/
				*buffer |= pad_value;	/*join them up*/

				/*Do the same encoding as before*/
				crypt_it(buffer, decode_option);

				/*now output the data we worked on*/
				count=fwrite(buffer, 1, BLOCK_SIZE, output);
				if (count != BLOCK_SIZE)
					bail("Error in writing to output stream-last block\n");
			}
		}

		/*
		* Copy the current block into the previous block
		* so it isn't overwritten
		*/
		memcpy (previous_block, buffer, BLOCK_SIZE);
	}
	
/********************************************************************************/
/*Close file handlers to ensure all data is flushed*/
if (input != stdin) fclose(input);
if (output != stdout) fclose(output);

/********************************************************************************/
/*Free up resources*/
/*Not really needed as this is the end of the program, but it's good practise*/
free(padded_keystring);
free(K);
free(S);

return (0);
}/*end of Main*/

/********************************************************************************/
/*Crypt or decrypt the buffer sent to us*/
void crypt_it(char *buffer, BOOLEAN decode)
{
ULLONG *L,*R;		/*pointers to parts of the buffer-left and right ULLONG chunks*/
int i;

	/*now we have sixteen bytes ready to work on*/
	/*split it into two ULLONGs, left and right*/

	L = (ULLONG *)buffer;
	R = (ULLONG *)(buffer+8);

	/*Put the data into host byte order*/
	*L = ntohll (*L);
	*R = ntohll (*R);

	if (decode)
	{
		/*decrypt the data...*/
		for (i=rounds;i>=1;i--)
		{
			*R = rrotate(*R - ntohll(S[2*i+1]) , *L ) ^ *L;
			*L = rrotate(*L - ntohll(S[2*i]), *R) ^ *R;
		}
		*L -= ntohll(S[0]);
		*R -= ntohll(S[1]);

	}
	else
	{
		/*run the data through the encryption the right number of times*/
		*L += ntohll(S[0]);
		*R += ntohll(S[1]);
		for (i=1;i<=rounds;i++)
		{
			*L = ntohll(S[2*i]) + lrotate( *L ^ *R, *R);
			*R = ntohll(S[2*i+1]) + lrotate(*R ^ *L, *L);
		}
	}

	/*Convert buffer back to network order*/
	*L = htonll(*L);
	*R = htonll(*R);
	
	return;
}

